import asyncio

from .stateful_cell import StatefulCell
from .observer_state import ObserverCellState
from .exceptions import UninitializedCellError, PendingAsyncValueError
from .extension import cell_extension
from .async_state import AsyncCellState
from .computed import computed
from .maybe import Maybe
from .keys import ValueKey

class WaitCell(StatefulCell):
    """A cell that *awaits* an *awaitable* held in the argument cell.

    The value of this cell is the completed value of the *awaitable*
    held in the argument cell. If the *awaitable*, raises an
    exception, then accessing this cell raises the same exception.

    .. important::

       If the value of this cell is accessed before the *awaitable*
       held in the argument cell has completed, a
       :any:`PendingAsyncValueError` exception is raised.

    .. note::

       The difference between this cell and ``AwaitCell`` is that the
       value of this cell is not reset when the value of the argument
       cell changes. Instead, the completed value of the last
       *awaitable* is kept until the new *awaitable* completes.

    :param arg: The argument cell, which holds an *awaitable*
    :type arg: Cell

    :param last_only: If True, only the last *awaitable* assigned to
                      the argument cell is awaited. If False, the cell
                      waits for every *awaitable* assigned to the
                      argument cell to complete.

    :type last_only: bool

    :param key: Key identifying the cell

    """

    def __init__(self, arg, last_only=False, key=None):
        super().__init__(key=key)

        self.arg = arg
        self.last_only = last_only

    @property
    def value(self):
        state = self.state

        if state is not None:
            return state.value

        raise UninitializedCellError

    def create_state(self):
        return WaitCellState(
            self,
            key=self.key
        )

class WaitCellState(AsyncCellState, ObserverCellState):
    """The state for a ``WaitCell``"""

    def __init__(self, cell, key):
        super().__init__(
            cell=cell,
            key=key
        )

        self.arg = cell.arg
        self.last_only = cell.last_only

    def on_will_update(self):
        # Prevent observers from being notified before awaitable
        # completion

        pass

    def on_update(self, did_change):
        # Prevent observers from being notified before awaitable
        # completion

        pass



import asyncio
import pytest

from asyncio import Future

from live_cells import (
    value,
    mutable,
    computed,
    wait,
    waited,
    batch,
    PendingAsyncValueError
)

from util import (
    observe,
    ValueTestObserver,
    MockException,
    future_value,
    future_error,
    delayed,
    is_pending
)

@pytest.mark.asyncio
class TestWaitCell:
    """Tests the behaviour of ``waited()`` cells."""

    async def test_constant_cell(self):
        """Test a simple coroutine held in a constant cell."""

        cell = value(future_value(12)).waited()

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            assert cell.value == 12
            assert observer.values == [12]

    async def test_mutable_cell(self):
        """Test a simple coroutine held in a mutable cell."""

        m = mutable(future_value(12))
        c = m.waited()

        with observe(c, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert c.value == 12

            m.value = future_value(100)

            is_pending(c)

            await asyncio.sleep(1)
            assert c.value == 100

            assert observer.values == [12, 100]

    async def test_notifies_observers(self):
        """Test that the observers of the cell are notified correctly."""

        m = mutable(future_value(12))
        c = m.waited()

        with observe(c, ValueTestObserver()) as observer:
            m.value = future_value(100)
            m.value = future_value(20)
            m.value = future_value(30)

            await asyncio.sleep(1)

            assert observer.values == [30]

    async def test_computed_cell(self):
        """Test waited on a computed cell that returns a coroutine."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        async def sum():
            x, y = await asyncio.gather(a(), b())
            return x + y

        cell = sum.waited()

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert cell.value == 3

            a.value = future_value(5)

            is_pending(cell)

            await asyncio.sleep(1)
            assert cell.value == 7

            b.value = future_value(10)

            is_pending(cell)

            await asyncio.sleep(1)
            assert cell.value == 15

            with batch():
                a.value = future_value(20)
                b.value = future_value(30)

            is_pending(cell)

            await asyncio.sleep(1)
            assert cell.value == 50

            assert observer.values == [3, 7, 15, 50]

    async def test_delayed_computed_cell(self):
        """Test waited() on computed cell with varying delays."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        async def sum():
            x, y = await asyncio.gather(a(), b())
            return x + y

        cell = sum.waited()

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert cell.value == 3

            with batch():
                a.value = future_value(20)
                b.value = delayed(10, 30)

            is_pending(cell)

            await asyncio.sleep(5)
            is_pending(cell)

            await asyncio.sleep(6)

            assert cell.value == 50
            assert observer.values == [3, 50]

    async def test_latest_value(self):
        """Test that only the latest coroutine is awaited even with varying delays."""

        f = mutable(delayed(10, 1))
        w = f.waited()

        with observe(w, ValueTestObserver()) as observer:
            f.value = future_value(2)
            f.value = delayed(30, 3)
            f.value = future_value(4)

            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(5)
            assert w.value == 4
            assert observer.values == [4]

            await asyncio.sleep(6)
            assert w.value == 4
            assert observer.values == [4]

            await asyncio.sleep(10)
            assert w.value == 4
            assert observer.values == [4]

            await asyncio.sleep(30)
            assert w.value == 4
            assert observer.values == [4]

            f.value = future_value(100)
            await asyncio.sleep(1)

            assert w.value == 100
            assert observer.values == [4, 100]

    async def test_exceptions(self):
        """Test that exceptions raised in coroutines are raised by waited() cells."""

        f = mutable(future_value(1))
        w = f.waited()

        with observe(w, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            f.value = future_error(MockException())
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            f.value = future_value(10)
            await asyncio.sleep(1)

            assert observer.values == [1, 10]

    async def test_compares_eq_same_cells(self):
        """Test that cells compare equal when given the same argument cell."""

        f = mutable(future_value(1))

        w1 = f.waited()
        w2 = f.waited()

        assert w1 == w2
        assert hash(w1) == hash(w2)

    async def test_compares_neq_different_cells(self):
        """Test that cells compare not equal when given different argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(1))

        w1 = f1.waited()
        w2 = f2.waited()

        assert w1 != w2
        assert w1 == w1


@pytest.mark.asyncio
class TestMultiWaitCell:
    """Test the behaviour of ``waited()`` with multiple arguments."""

    async def test_constant_cells(self):
        """Test simple coroutines held in constant cells."""

        a = value(future_value(1))
        b = value(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

    async def test_mutable_cells(self):
        """Test simple coroutines held in mutable cells."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

            a.value = future_value(5)
            is_pending(sum)

            await asyncio.sleep(1)
            assert sum.value == 7

            b.value = future_value(10)
            is_pending(sum)

            await asyncio.sleep(1)
            assert sum.value == 15

            with batch():
                a.value = future_value(20)
                b.value = future_value(30)

            is_pending(sum)

            await asyncio.sleep(1)
            assert sum.value == 50

            assert observer.values == [3, 7, 15, 50]

    async def test_notifies_observers(self):
        """Test that the observers of the cell are notified correctly."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            a.value = future_value(15)
            b.value = future_value(20)

            with batch():
                a.value = future_value(100)
                b.value = future_value(320)

            await asyncio.sleep(1)
            assert observer.values == [420]

    async def test_delays(self):
        """Test that coroutines with delays are handled correctly."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

            with batch():
                a.value = future_value(20)
                b.value = delayed(10, 30)

            is_pending(sum)

            await asyncio.sleep(5)
            is_pending(sum)

            await asyncio.sleep(6)
            assert sum.value == 50

            assert observer.values == [3, 50]

    async def test_latest_value(self):
        """Test that only the latest coroutine is awaited even with varying delays."""

        a = mutable(delayed(10, 1))
        b = mutable(future_value(2))

        @computed
        def w():
            x, y = wait(a, b)
            return x + y

        with observe(w, ValueTestObserver()) as observer:
            a.value = future_value(10)

            with batch():
                a.value = delayed(30, 20)
                b.value = future_value(7)

            a.value = future_value(100)

            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(5)
            assert w.value == 107
            assert observer.values == [107]

            await asyncio.sleep(6)
            assert w.value == 107
            assert observer.values == [107]

            await asyncio.sleep(10)
            assert w.value == 107
            assert observer.values == [107]

            await asyncio.sleep(10)
            assert w.value == 107
            assert observer.values == [107]

            a.value = future_value(1000)
            await asyncio.sleep(1)
            assert w.value == 1007
            assert observer.values == [107, 1007]

    async def test_exceptions(self):
        """Test that exceptions raised in coroutines are raised by awaited() cells."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))
        w = waited(a, b)

        with observe(w, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            a.value = future_error(MockException())
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            a.value = future_value(10)
            b.value = future_error(MockException)
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            b.value = future_value(15)
            await asyncio.sleep(1)

            assert observer.values == [[1, 2], [10, 15]]

    async def test_compares_eq_same_cells(self):
        """Test that cells compare equal when given the same argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(3))

        w1 = waited(f1, f2)
        w2 = waited(f1, f2)

        assert w1 == w2
        assert hash(w1) == hash(w2)

    async def test_compares_neq_difference_cells(self):
        """Test that cells compare not equal when given different argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(3))
        f3 = mutable(future_value(5))

        w1 = waited(f1, f2)
        w2 = waited(f1, f3)
        w3 = waited(f3, f2)

        assert w1 == w1
        assert w1 != w2
        assert w1 != w3
        assert w2 != w3

@pytest.mark.asyncio
class TestWaitNoResetCell:
    """Tests the behaviour of ``waited()`` cells with reset=False"""

    async def test_constant_cell(self):
        """Test a simple coroutine held in a constant cell."""

        cell = value(future_value(12)).waited(reset=False)

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            assert cell.value == 12
            assert observer.values == [12]

    async def test_mutable_cell(self):
        """Test a simple coroutine held in a mutable cell."""

        m = mutable(future_value(12))
        c = m.waited(reset=False)

        with observe(c, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert c.value == 12

            m.value = future_value(100)
            assert c.value == 12

            await asyncio.sleep(1)
            assert c.value == 100

            assert observer.values == [12, 100]

    async def test_notifies_observers(self):
        """Test that the observers of the cell are notified correctly."""

        m = mutable(future_value(12))
        c = m.waited(reset=False)

        with observe(c, ValueTestObserver()) as observer:
            m.value = future_value(100)
            m.value = future_value(20)
            m.value = future_value(30)

            await asyncio.sleep(1)

            assert observer.values == [30]

    async def test_computed_cell(self):
        """Test waited on a computed cell that returns a coroutine."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        async def sum():
            x, y = await asyncio.gather(a(), b())
            return x + y

        cell = sum.waited(reset=False)

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert cell.value == 3

            a.value = future_value(5)
            assert cell.value == 3

            await asyncio.sleep(1)
            assert cell.value == 7

            b.value = future_value(10)
            assert cell.value == 7

            await asyncio.sleep(1)
            assert cell.value == 15

            with batch():
                a.value = future_value(20)
                b.value = future_value(30)

            assert cell.value == 15

            await asyncio.sleep(1)
            assert cell.value == 50

            assert observer.values == [3, 7, 15, 50]

    async def test_delayed_computed_cell(self):
        """Test waited() on computed cell with varying delays."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        async def sum():
            x, y = await asyncio.gather(a(), b())
            return x + y

        cell = sum.waited(reset=False)

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert cell.value == 3

            with batch():
                a.value = future_value(20)
                b.value = delayed(10, 30)

            assert cell.value == 3

            await asyncio.sleep(5)
            assert cell.value == 3

            await asyncio.sleep(6)

            assert cell.value == 50
            assert observer.values == [3, 50]

    async def test_latest_value(self):
        """Test that only the latest coroutine is awaited even with varying delays."""

        f = mutable(delayed(10, 1))
        w = f.waited(reset=False)

        with observe(w, ValueTestObserver()) as observer:
            f.value = future_value(2)
            f.value = delayed(30, 3)
            f.value = future_value(4)

            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(5)
            assert w.value == 4
            assert observer.values == [4]

            await asyncio.sleep(6)
            assert w.value == 4
            assert observer.values == [4]

            await asyncio.sleep(10)
            assert w.value == 4
            assert observer.values == [4]

            await asyncio.sleep(30)
            assert w.value == 4
            assert observer.values == [4]

            f.value = future_value(100)
            await asyncio.sleep(1)

            assert w.value == 100
            assert observer.values == [4, 100]

    async def test_exceptions(self):
        """Test that exceptions raised in coroutines are raised by waited() cells."""

        f = mutable(future_value(1))
        w = f.waited(reset=False)

        with observe(w, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            f.value = future_error(MockException())
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            f.value = future_value(10)
            await asyncio.sleep(1)

            assert observer.values == [1, 10]

    async def test_compares_eq_same_cells(self):
        """Test that cells compare equal when given the same argument cell."""

        f = mutable(future_value(1))

        w1 = f.waited(reset=False)
        w2 = f.waited(reset=False)

        assert w1 == w2
        assert hash(w1) == hash(w2)

    async def test_compares_neq_different_cells(self):
        """Test that cells compare not equal when given different argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(1))

        w1 = f1.waited(reset=False)
        w2 = f2.waited(reset=False)

        assert w1 != w2
        assert w1 == w1

    async def test_compares_neq_different_kwargs(self):
        """Test that cells compare not equal when given different keyword arguments."""

        f = mutable(future_value(1))

        w1 = f.waited()
        w2 = f.waited(reset=False)
        w3 = f.waited(reset=False, queue=True)

        assert w1 == w1
        assert w1 != w2
        assert w1 != w3

        assert w2 == w2
        assert w2 != w1
        assert w2 != w3

        assert w3 == w3
        assert w3 != w1
        assert w3 != w2

@pytest.mark.asyncio
class TestMultiWaitNoResetCell:
    """Test behaviour of ``waited(reset=False)`` with multiple arguments."""

    async def test_constant_cells(self):
        """Test simple coroutines held in constant cells."""

        a = value(future_value(1))
        b = value(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

    async def test_mutable_cells(self):
        """Test simple coroutines held in mutable cells."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

            a.value = future_value(5)
            assert sum.value == 3

            await asyncio.sleep(1)
            assert sum.value == 7

            b.value = future_value(10)
            assert sum.value == 7

            await asyncio.sleep(1)
            assert sum.value == 15

            with batch():
                a.value = future_value(20)
                b.value = future_value(30)

            assert sum.value == 15

            await asyncio.sleep(1)
            assert sum.value == 50

            assert observer.values == [3, 7, 15, 50]

    async def test_notifies_observers(self):
        """Test that the observers of the cell are notified correctly."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            a.value = future_value(15)
            b.value = future_value(20)

            with batch():
                a.value = future_value(100)
                b.value = future_value(320)

            await asyncio.sleep(1)
            assert observer.values == [420]

    async def test_delays(self):
        """Test that coroutines with delays are handled correctly."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

            with batch():
                a.value = future_value(20)
                b.value = delayed(10, 30)

            assert sum.value == 3

            await asyncio.sleep(5)
            assert sum.value == 3

            await asyncio.sleep(6)
            assert sum.value == 50

            assert observer.values == [3, 50]

    async def test_latest_value(self):
        """Test that only the latest coroutine is awaited even with varying delays."""

        a = mutable(delayed(10, 1))
        b = mutable(future_value(2))

        @computed
        def w():
            x, y = wait(a, b, reset=False)
            return x + y

        with observe(w, ValueTestObserver()) as observer:
            a.value = future_value(10)

            with batch():
                a.value = delayed(30, 20)
                b.value = future_value(7)

            a.value = future_value(100)

            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(5)
            assert w.value == 107
            assert observer.values == [107]

            await asyncio.sleep(6)
            assert w.value == 107
            assert observer.values == [107]

            await asyncio.sleep(10)
            assert w.value == 107
            assert observer.values == [107]

            await asyncio.sleep(10)
            assert w.value == 107
            assert observer.values == [107]

            a.value = future_value(1000)
            await asyncio.sleep(1)
            assert w.value == 1007
            assert observer.values == [107, 1007]

    async def test_exceptions(self):
        """Test that exceptions raised in coroutines are raised by awaited() cells."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))
        w = waited(a, b, reset=False)

        with observe(w, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            a.value = future_error(MockException())
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            a.value = future_value(10)
            b.value = future_error(MockException)
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            b.value = future_value(15)
            await asyncio.sleep(1)

            assert observer.values == [[1, 2], [10, 15]]

    async def test_compares_eq_same_cells(self):
        """Test that cells compare equal when given the same argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(3))

        w1 = waited(f1, f2, reset=False)
        w2 = waited(f1, f2, reset=False)

        assert w1 == w2
        assert hash(w1) == hash(w2)

    async def test_compares_neq_difference_cells(self):
        """Test that cells compare not equal when given different argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(3))
        f3 = mutable(future_value(5))

        w1 = waited(f1, f2, reset=False)
        w2 = waited(f1, f3, reset=False)
        w3 = waited(f3, f2, reset=False)

        assert w1 == w1
        assert w1 != w2
        assert w1 != w3
        assert w2 != w3

    async def test_compares_neq_different_kwargs(self):
        """Test that cells compare not equal when given different keyword arguments."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(2))

        w1 = waited(f1, f2)
        w2 = waited(f1, f2, reset=False)
        w3 = waited(f1, f2, reset=False, queue=True)

        assert w1 == w1
        assert w1 != w2
        assert w1 != w3

        assert w2 == w2
        assert w2 != w1
        assert w2 != w3

        assert w3 == w3
        assert w3 != w1
        assert w3 != w2


@pytest.mark.asyncio
class TestWaitQueueCell:
    """Tests the behaviour of ``waited()`` cells with queue=True"""

    async def test_constant_cell(self):
        """Test a simple coroutine held in a constant cell."""

        cell = value(future_value(12)).waited(reset=False, queue=True)

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            assert cell.value == 12
            assert observer.values == [12]

    async def test_mutable_cell(self):
        """Test a simple coroutine held in a mutable cell."""

        m = mutable(future_value(12))
        c = m.waited(reset=False, queue=True)

        with observe(c, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert c.value == 12

            m.value = future_value(100)
            assert c.value == 12

            await asyncio.sleep(1)
            assert c.value == 100

            assert observer.values == [12, 100]

    async def test_notifies_observers(self):
        """Test that the observers of the cell are notified correctly."""

        m = mutable(future_value(12))
        c = m.waited(reset=False, queue=True)

        with observe(c, ValueTestObserver()) as observer:
            m.value = future_value(100)
            m.value = future_value(20)
            m.value = future_value(30)

            await asyncio.sleep(1)

            assert observer.values == [12, 100, 20, 30]

    async def test_computed_cell(self):
        """Test waited on a computed cell that returns a coroutine."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        async def sum():
            x, y = await asyncio.gather(a(), b())
            return x + y

        cell = sum.waited(reset=False, queue=True)

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert cell.value == 3

            a.value = future_value(5)
            assert cell.value == 3

            await asyncio.sleep(1)
            assert cell.value == 7

            b.value = future_value(10)
            assert cell.value == 7

            await asyncio.sleep(1)
            assert cell.value == 15

            with batch():
                a.value = future_value(20)
                b.value = future_value(30)

            assert cell.value == 15

            await asyncio.sleep(1)
            assert cell.value == 50

            assert observer.values == [3, 7, 15, 50]

    async def test_delayed_computed_cell(self):
        """Test waited() on computed cell with varying delays."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        async def sum():
            x, y = await asyncio.gather(a(), b())
            return x + y

        cell = sum.waited(reset=False, queue=True)

        with observe(cell, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert cell.value == 3

            with batch():
                a.value = future_value(20)
                b.value = delayed(10, 30)

            assert cell.value == 3

            await asyncio.sleep(5)
            assert cell.value == 3

            await asyncio.sleep(6)

            assert cell.value == 50
            assert observer.values == [3, 50]

    async def test_queued_values(self):
        """Test all coroutines are awaited correctly even with varying delays."""

        f = mutable(delayed(10, 1))
        w = f.waited(reset=False, queue=True)

        with observe(w, ValueTestObserver()) as observer:
            f.value = future_value(2)
            f.value = delayed(30, 3)
            f.value = future_value(4)

            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(5)
            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(6)
            assert w.value == 2
            assert observer.values == [1, 2]

            await asyncio.sleep(10)
            assert w.value == 2
            assert observer.values == [1, 2]

            await asyncio.sleep(30)
            assert w.value == 4
            assert observer.values == [1, 2, 3, 4]

            f.value = future_value(100)
            await asyncio.sleep(1)

            assert w.value == 100
            assert observer.values == [1, 2, 3, 4, 100]

    async def test_exceptions(self):
        """Test that exceptions raised in coroutines are raised by waited() cells."""

        f = mutable(future_value(1))
        w = f.waited(reset=False, queue=True)

        with observe(w, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            f.value = future_error(MockException())
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            f.value = future_value(10)
            await asyncio.sleep(1)

            assert observer.values == [1, 10]

    async def test_compares_eq_same_cells(self):
        """Test that cells compare equal when given the same argument cell."""

        f = mutable(future_value(1))

        w1 = f.waited(reset=False, queue=True)
        w2 = f.waited(reset=False, queue=True)

        assert w1 == w2
        assert hash(w1) == hash(w2)

    async def test_compares_neq_different_cells(self):
        """Test that cells compare not equal when given different argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(1))

        w1 = f1.waited(reset=False, queue=True)
        w2 = f2.waited(reset=False, queue=True)

        assert w1 != w2
        assert w1 == w1

@pytest.mark.asyncio
class TestMultiWaitQueueCell:
    """Test behaviour of ``waited(queue=True)`` with multiple arguments."""

    async def test_constant_cells(self):
        """Test simple coroutines held in constant cells."""

        a = value(future_value(1))
        b = value(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False, queue=True)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

    async def test_mutable_cells(self):
        """Test simple coroutines held in mutable cells."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False, queue=True)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

            a.value = future_value(5)
            assert sum.value == 3

            await asyncio.sleep(1)
            assert sum.value == 7

            b.value = future_value(10)
            assert sum.value == 7

            await asyncio.sleep(1)
            assert sum.value == 15

            with batch():
                a.value = future_value(20)
                b.value = future_value(30)

            assert sum.value == 15

            await asyncio.sleep(1)
            assert sum.value == 50

            assert observer.values == [3, 7, 15, 50]

    async def test_notifies_observers(self):
        """Test that the observers of the cell are notified correctly."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False, queue=True)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            a.value = future_value(15)
            b.value = future_value(20)

            with batch():
                a.value = future_value(100)
                b.value = future_value(320)

            await asyncio.sleep(1)
            assert observer.values == [3, 17, 35, 420]

    async def test_delays(self):
        """Test that coroutines with delays are handled correctly."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))

        @computed
        def sum():
            x, y = wait(a, b, reset=False, queue=True)
            return x + y

        with observe(sum, ValueTestObserver()) as observer:
            await asyncio.sleep(1)
            assert sum.value == 3

            with batch():
                a.value = future_value(20)
                b.value = delayed(10, 30)

            assert sum.value == 3

            await asyncio.sleep(5)
            assert sum.value == 3

            await asyncio.sleep(6)
            assert sum.value == 50

            assert observer.values == [3, 50]

    async def test_queued_value(self):
        """Test all coroutines are awaited correctly even with varying delays."""

        a = mutable(delayed(10, 1))
        b = mutable(future_value(2))

        @computed
        def w():
            x, y = wait(a, b, reset=False, queue=True)
            return x + y

        with observe(w, ValueTestObserver()) as observer:
            a.value = future_value(10)

            with batch():
                a.value = delayed(30, 20)
                b.value = future_value(7)

            a.value = future_value(100)

            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(5)
            is_pending(w)
            assert observer.values == []

            await asyncio.sleep(6)
            assert w.value == 12
            assert observer.values == [3, 12]

            await asyncio.sleep(10)
            assert w.value == 12
            assert observer.values == [3, 12]

            await asyncio.sleep(20)
            assert w.value == 107
            assert observer.values == [3, 12, 27, 107]

            a.value = future_value(1000)
            await asyncio.sleep(1)
            assert w.value == 1007
            assert observer.values == [3, 12, 27, 107, 1007]

    async def test_exceptions(self):
        """Test that exceptions raised in coroutines are raised by awaited() cells."""

        a = mutable(future_value(1))
        b = mutable(future_value(2))
        w = waited(a, b, reset=False, queue=True)

        with observe(w, ValueTestObserver()) as observer:
            await asyncio.sleep(1)

            a.value = future_error(MockException())
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            a.value = future_value(10)
            b.value = future_error(MockException)
            await asyncio.sleep(1)

            with pytest.raises(MockException):
                w.value

            b.value = future_value(15)
            await asyncio.sleep(1)

            assert observer.values == [[1, 2], [10, 2], [10, 15]]

    async def test_compares_eq_same_cells(self):
        """Test that cells compare equal when given the same argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(3))

        w1 = waited(f1, f2, reset=False, queue=True)
        w2 = waited(f1, f2, reset=False, queue=True)

        assert w1 == w2
        assert hash(w1) == hash(w2)

    async def test_compares_neq_difference_cells(self):
        """Test that cells compare not equal when given different argument cells."""

        f1 = mutable(future_value(1))
        f2 = mutable(future_value(3))
        f3 = mutable(future_value(5))

        w1 = waited(f1, f2, reset=False, queue=True)
        w2 = waited(f1, f3, reset=False, queue=True)
        w3 = waited(f3, f2, reset=False, queue=True)

        assert w1 == w1
        assert w1 != w2
        assert w1 != w3
        assert w2 != w3



import asyncio

from .keys import ValueKey
from .computed import computed
from .extension import cell_extension
from .await_cell import AwaitCell
from .wait_cell import WaitCell

class WaitedCellKey(ValueKey):
    """Key identifying cells created with ``waited()``."""

@cell_extension
def waited(self, *cells, **kwargs):
    """Create a *wait cell* that *awaits* one or more *asynchronous cells*.

    The value of the returned cell is the completed value of the
    *awaitable* held in ``self``. If the *awaitable*, or the cell
    ``self`` itself, raises an exception then accessing the value of
    the returned cell raises the same exception.

    .. important::

       If the value of the returned cell is accessed before the
       coroutine held in ``self`` has completed, a
       :any:`PendingAsyncValueError` exception is raised.


    If multiple arguments are given, a cell is returned that *awaits* the
    *awaitables* held in ``self`` and each cell in ``cells``. The value
    of the cell is a list holding the completed value of ``self``
    followed by the completed value of each cell in ``cells``. This is
    equivalent to the following:

    .. code-block:: python

       [await self(), await cells[0](), ...]

    The difference between a single call to ``waited``:

    .. code-block:: python

       @computed
       def sum():
           x, y = waited(a, b)()
           return x, y

    and multiple calls to ``waited``:

    .. code-block:: python

       @computed
       def sum():
           return a.waited()() + b.waited()()

    becomes apparent when an update to the *asynchronous cells* is
    triggered by a common ancestor of the cells. With a single call to
    ``waited``, the value of the ``sum`` cell is recomputed only once
    when both the *awaitables* held in ``a`` and ``b`` have
    completed. In the second example, the value of ``sum`` is
    recomputed once when the *awaitable* held in ``a`` has completed
    and a second time when the *awaitable* held in ``b`` has
    completed.

    This function also accepts a number of keyword arguments that
    modify the behaviour of the cell.

    The ``reset`` keyword argument controls whether the value of the
    cell is reset when the values of the asynchronous cells
    changes.

    By default this is ``True``, which means that when the values of
    the asynchronous cells changes the value of the returned *wait
    cell* is reset. Accessing the value of the returned cell at this
    point results in a :any:`PendingAsyncValueError` exception being
    raised, until the *awaitables* held in the asynchronous cells have
    completed.

    .. code-block:: python

       async def coro1():
           await asyncio.sleep(1)
           return 1

       async def coro2():
           await asyncio.sleep(1)
           return 2

       arg = mutable(coro1)

       w1 = arg.waited()
       w2 = arg.waited(reset=False)

       await asyncio.sleep(2)

       # The values of both `w1` and `w2` are 1

       arg.value = coro2

       # Accessing the value of `w1` raises `PendingAsyncValueError`
       # The value of `w2` is still 1


    The ``queue`` keyword argument controls what happens when the
    values of the asynchronous cells change before the *awaitables*,
    for which the *wait cell* is currently waiting, have completed.

    If ``queue`` is True, when the values of the asynchronous cells
    change, the *wait cell* first waits for the previous *awaitables*
    to complete, and only then waits for the new *awaitables*. The
    value of the *wait cell* is updated in the order in which the
    *awaitables* were assigned to the asynchronous cell and not
    necessarily in the order of completion.

    If ``queue`` is False, the default, the *wait cell* only waits for
    the last *awaitable* and ignores previous *awaitables* that have
    not completed.

    The following example illustrates the difference:

    .. code-block:: python

       async def coro1():
           await asyncio.sleep(5)
           return 1

       async def coro2():
           await asyncio.sleep(1)
           return 2

       arg = mutable(coro1)

       w1 = arg.waited(reset=False, queue=True)
       w2 = arg.waited(reset=False)

       @watch
       def watch_w1():
           print(f'W1 = {w1()}')

       @watch
       def watch_w2():
           print(f'W2 = {w2()}')

       # NOTE: There is no await meaning coro1 has not completed
       arg.value = coro2


    For ``w1`` two lines are printed: ``W1 = 1`` and ``W1 = 2``, while
    for ``w2`` only a single line ``W2 = 2`` is printed.

    .. attention::

       The ``queue`` argument has no effect if ``reset`` is
       True.

    :param cells: Additional cells to ``await``
    :type cells: List[Cell]

    The following keyword arguments are accepted:

    :param reset: If True (the default), the value of the cell is
                  reset, whenever the values of the asynchronous cells
                  change. Otherwise the last completed value (or
                  exception) is retained until the new *awaitables* have
                  completed.

    :type reset: bool

    :param queue: If true, the returned cell waits for all
                  `awaitables` that were assigned to the asynchronous
                  cell(s) to complete. If false, the cell only waits
                  for the last *awaitables* that were assigned.

    :type queue: bool

    :returns: The *wait cell*
    :rtype: Cell

    """

    reset = kwargs.get('reset', True)
    queue = kwargs.get('queue', False)

    arg = self

    if cells:
        @computed
        def gathered():
            return asyncio.gather(
                self(),
                *(c() for c in cells)
            )

        arg = gathered

    key = WaitedCellKey(reset, queue, self, *cells)

    if reset:
        return AwaitCell(arg, key=key)

    else:
        return WaitCell(
            arg,
            last_only=not queue,
            key=key
        )

@cell_extension
def wait(self, *cells, **kwargs):
    """Create a *wait cell*, as if by :any:`waited` and access its value.

    Calling this function is equivalent to:

    .. code-block:: python

       waited(self, *cells, **kwargs)()

    This function accepts the same positional and keyword arguments as
    :any:`waited`.

    :returns: The value of the *wait cell*

    """

    return waited(self, *cells, **kwargs)()



from .stateful_cell import StatefulCell
from .observer_state import ObserverCellState
from .exceptions import UninitializedCellError, PendingAsyncValueError
from .async_state import AsyncCellState
from .maybe import Maybe

class AwaitCell(StatefulCell):
    """A cell that *awaits* an *awaitable* held in the argument cell.

    The value of this cell is the completed value of the *awaitable*
    held in the argument cell. If the *awaitable*, raises an exception,
    then accessing this cell raises the same exception.

    .. important::

       If the value of this cell is accessed before the *awaitable*
       held in the cell has completed, a :any:`PendingAsyncValueError`
       exception is raised.

    :param arg: The argument cell, which holds an *awaitable*
    :type arg: Cell

    :param key: Key identifying the cell

    """

    def __init__(self, arg, key=None):
        super().__init__(key=key)

        self.arg = arg

    @property
    def value(self):
        state = self.state

        if state is not None:
            return state.value

        raise UninitializedCellError


    def create_state(self):
        return AwaitCellState(self, key=self.key)

class AwaitCellState(AsyncCellState, ObserverCellState):
    """Implements the cell state for an ``AwaitCell``"""

    def __init__(self, cell, key):
        super().__init__(
            cell = cell,
            key = key
        )

        self.arg = cell.arg
        self.last_only = True

    def on_update(self, did_change):
        if did_change:
            self._awaited_value = Maybe(
                error=PendingAsyncValueError()
            )

        super().on_update(did_change)



import asyncio
from asyncio import Future

from .observer_state import ObserverCellState
from .exceptions import PendingAsyncValueError
from .maybe import Maybe

class AsyncCellState:
    """Augments ``CellState`` with functionality for waiting for *awaitables*.

    This mixin provides functionality for creating a ``CellState``
    that updates its value when the *awaitable* held in the argument
    cell completes.

    The behaviour of this class is influenced by the value of the
    ``last_only`` property.

    If ``last_only`` is false, the state waits for the completion of
    every *awaitable* that is held in the argument cell. The values
    assigned to the cell associated with this state, are the completed
    values of the *awaitables* held in the argument cell. The
    assignments happen in the order, in which the *awaitables* are
    assigned to the argument cell and not necessarily in the order of
    completion of the *awaitables*.

    If ``last_only`` is true and the argument cell is assigned a new
    *awaitable* before the previous *awaitable* completes, the state only
    waits for the new *awaitable* to complete, and does not emit an
    update for the previous *awaitable*.

    """

    def init(self):
        super().init()

        self._task = None
        self._awaited_value = Maybe(error=PendingAsyncValueError())

        self.arg.add_observer(self)
        self._update_value(self._get_value())

    def dispose(self):
        self.arg.remove_observer(self)

        if self._task is not None:
            self._task.cancel()

        super().dispose()

    @property
    def value(self):
        return self._awaited_value.unwrap()

    def post_update(self):
        super().post_update()

        if self.last_only and self._task is not None:
            self._task.cancel()
            self._task = None

        self._update_value(self._get_value())

    def _get_value(self):
        return Maybe.wrap_async(lambda: self.arg.value)

    def _update_value(self, future_value):
        self._task = asyncio.create_task(
            self.wait_future(self._task, future_value)
        )

    async def wait_future(self, task, future_value):
        if task is not None:
            await asyncio.shield(task)

        self._set_value(await future_value)

    def _set_value(self, value):
        self.notify_will_update()

        self.stale = False
        self._awaited_value = value

        self.notify_update()
